"""Spell checking system calls to Hunspell."""

import glob
import os
import subprocess
import tempfile


def hunspell_spellcheck(
    content,
    language_dicts,
    personal_dicts=None,
    encoding=None,
):
    """Call hunspell for spellchecing.

    Args:
        content (str): Content to check for words not included in dictionaries.
        language_dicts (list, str): Language or languages dictionaries (could
            be defined as files) used to check errors.
        personal_dicts (str): Personal dictionary used to exclude valid words
            from being notified as errors.
        encoding (str): Input encoding passed to Hunspell. If is defined, the
            option ``-i`` will be passed to hunspell system call.

    Returns:
        str: Hunspell standard output.
    """
    if not isinstance(language_dicts, str):
        language_dicts = ",".join(language_dicts)

    command = ["hunspell", "-d", language_dicts, "-a"]

    temporal_personal_dict_filename = None
    if personal_dicts:
        if isinstance(personal_dicts, str):
            command.extend(["-p", personal_dicts])
        else:
            if len(personal_dicts) == 1:
                # only one dictionary, no need for composition
                command.extend(["-p", personal_dicts[0]])
            else:
                # compound one dictionary with others content
                hunspellcheck_tempdir = os.path.join(
                    tempfile.gettempdir(),
                    "hunspellcheck",
                )
                if not os.path.isdir(hunspellcheck_tempdir):
                    os.mkdir(hunspellcheck_tempdir)

                temporal_personal_dict_filename = tempfile.NamedTemporaryFile(
                    dir=hunspellcheck_tempdir
                ).name
                if os.path.isfile(temporal_personal_dict_filename):  # pragma: no cover
                    os.remove(temporal_personal_dict_filename)
                os.mknod(temporal_personal_dict_filename)
                for personal_dict_glob in personal_dicts:
                    for personal_dict in glob.glob(personal_dict_glob):
                        with open(
                            temporal_personal_dict_filename, "a"
                        ) as compound_dict_f, open(personal_dict) as partial_dict_f:
                            compound_dict_f.write(f"{partial_dict_f.read()}\n")
                command.extend(["-p", temporal_personal_dict_filename])

    if encoding:
        command.extend(["-i"], encoding)

    response = subprocess.run(
        command,
        universal_newlines=True,
        input=content,
        stdout=subprocess.PIPE,
        check=True,
    )

    if temporal_personal_dict_filename is not None:
        os.remove(temporal_personal_dict_filename)

    return response



"""Files positional argument related stuff for hunspellcheck CLI utilities."""

import argparse
import copy
import glob


class FilesOrGlobsAction(argparse.Action):
    """Prior to Python3.8, the argarse module does not include the
    `_ExtendAction`, so here we are replicating their behaviour.

    If the library stops supporting Python < 3.8, might be useful simplify
    this code using `_ExtendAction` with a super call passing as the extended
    filenames as values.
    """

    def __call__(self, parser, namespace, values, option_string=None):
        items = copy.copy(getattr(namespace, self.dest, []) or [])
        for value in values:
            items.extend(glob.glob(value))
        setattr(namespace, self.dest, items)



"""Version CLI option utilities for hunspellcheck."""

from jinja2 import Template

from hunspellcheck.hunspell.version import get_hunspell_version


DEFAULT_VERSION_TEMPLATE = (
    "{% if version_number %}{{version_prog}} {{version_number}}{% endif %}"
    "{% if version_number and (hunspell_version or ispell_version) %} - {% endif %}"
    "{% if hunspell_version %}Hunspell {{hunspell_version}}{% endif %}"
    "{% if hunspell_version and ispell_version %} - {% endif %}"
    "{% if ispell_version %}Ispell {{ispell_version}}{% endif %}"
)


def render_version_template(
    version_template,
    version_template_context,
    version_prog=None,
    version_number=None,
    hunspell_version=True,
    ispell_version=True,
):
    versions = (
        {}
        if not hunspell_version and not ispell_version
        else get_hunspell_version(hunspell=hunspell_version, ispell=ispell_version)
    )

    _version_template_context = {
        "hunspell_version": None if not hunspell_version else versions["hunspell"],
        "ispell_version": None if not ispell_version else versions["ispell"],
    }
    if version_number is not None:
        _version_template_context["version_prog"] = version_prog
        _version_template_context["version_number"] = version_number
    _version_template_context.update(version_template_context)

    return Template(version_template).render(**_version_template_context)



"""Utilities related with Hunspell versions."""

import os
import re
import subprocess


def get_hunspell_version(hunspell=True, ispell=True):
    """Returns the number of version of Hunspell and the version of Ispell
    that the installed Hunspell program is using.

    Args:
        hunspell (bool): Include the version of Hunspell in the response.
        ispell (bool): Include the version of Ispell in the response.

    Returns:
        dict: Their fields would be `hunspell` and `ispell`, if both included
            using the kwargs of this function.
    """
    if not hunspell and not ispell:
        raise ValueError(
            "At least one of optional arguments 'hunspell' or 'ispell' must be true."
        )

    previous_env_lang = os.environ.get("LANG", None)
    os.environ["LANG"] = "C"

    output = subprocess.run(
        ["hunspell", "--version"],
        stdout=subprocess.PIPE,
        universal_newlines=True,
    )

    if previous_env_lang is None:
        del os.environ["LANG"]
    else:
        os.environ["LANG"] = previous_env_lang

    version_line = output.stdout.splitlines()[0]

    response = {}

    if hunspell:
        response["hunspell"] = re.search(
            r"Hunspell\s([^\n\s)]+)",
            version_line,
        ).group(1)

    if ispell:
        response["ispell"] = re.search(
            r"Ispell\s(Version)?\s([^\n\s]+)", version_line
        ).group(2)
    return response



"""Hunspellcheck package."""

from hunspellcheck.cli import hunspellchecker_argument_parser
from hunspellcheck.exceptions import InvalidLanguageDictionaryError
from hunspellcheck.hunspell.dictionaries import (
    assert_is_valid_dictionary_language_or_filename,
    gen_available_dictionaries,
    gen_available_dictionaries_with_langcodes,
    is_valid_dictionary_language,
    is_valid_dictionary_language_or_filename,
    list_available_dictionaries,
    print_available_dictionaries,
)
from hunspellcheck.hunspell.version import get_hunspell_version
from hunspellcheck.spellchecker import HunspellChecker, render_hunspell_word_error
from hunspellcheck.word import looks_like_a_word_creator


__version__ = "0.0.8"
__title__ = "hunspellcheck"
__all__ = (
    "HunspellChecker",
    "InvalidLanguageDictionaryError",
    "hunspellchecker_argument_parser",
    "gen_available_dictionaries",
    "gen_available_dictionaries_with_langcodes",
    "get_hunspell_version",
    "is_valid_dictionary_language",
    "is_valid_dictionary_language_or_filename",
    "assert_is_valid_dictionary_language_or_filename",
    "list_available_dictionaries",
    "looks_like_a_word_creator",
    "print_available_dictionaries",
    "render_hunspell_word_error",
)



"""Personal dictionary CLI option stuff."""

import argparse
import copy
import glob


class PersonalDictionaryAction(argparse.Action):
    def __call__(self, parser, namespace, values, option_string=None):
        items = copy.copy(getattr(namespace, self.dest, []) or [])
        for value in values:
            items.extend(glob.glob(value))
        setattr(namespace, self.dest, items)



"""Spellchecker of hunspellcheck.

This module contains all the spellchecking logic.
"""

from hunspellcheck.exceptions import Unreachable
from hunspellcheck.hunspell.spellcheck import hunspell_spellcheck
from hunspellcheck.word import looks_like_a_word_creator


ERROR_FIELDS = [
    "filename",
    "line_number",
    "word",
    "word_line_index",
    "line",
    "text",
    "error_number",
    "near_misses",
]

DEFAULT_LOOKS_LIKE_A_WORD = looks_like_a_word_creator()


class HunspellChecker:
    """Main spellchecking interface of hunspellcheck.

    Args:
        filenames_contents (dict): Dictionary mapping filenames to content of
            those files.
        languages (list, str): Languages against will be checked the contents.
        personal_dicts (str, list): Globs of files which would be dictionaries
            with custom words to ignore from being triggered as positives. Can
            be globs or files, as string or list of strings.
        looks_like_a_word (types.FunctionType): Function to filter the positive
            words from being considered positives. Takes a possible word string
            and returns if the value could be considered a word to be checked
            for mispelling errors. By default, the function
            :py:func:`hunspellcheck.word.looks_like_a_word_creator` will be
            used with all its arguments by default to build a basic validator.
        encoding (str): Input encoding. If not defined, it will be autodetected
            by hunspell.
    """

    def __init__(
        self,
        filenames_contents,
        languages,
        personal_dicts=None,
        looks_like_a_word=DEFAULT_LOOKS_LIKE_A_WORD,
        encoding=None,
    ):
        self.filenames_contents = filenames_contents
        self.languages = languages
        self.personal_dicts = personal_dicts
        self.looks_like_a_word = looks_like_a_word
        self.errors = None
        self.encoding = encoding

    def check(
        self,
        include_filename=True,
        include_line_number=True,
        include_word=True,
        include_word_line_index=True,
        include_line=False,
        include_text=False,
        include_error_number=False,
        include_near_misses=False,
    ):
        """Spellchecking function.

        Yields each mispelled word data found in contents from a generator. The
        data generated for each word depends on the optional arguments
        ``include_<field>`` passed to this function, being ``field`` the name
        of the field inside the yielded dictionary.

        Args:
            include_filename (bool): Includes filename where the mispelled word
                has been found in yielded error data.
            include_line_number (bool): Includes the line number where the
                mispelled word has been found in the content for the yielded
                error data.
            include_word (bool): Includes the mispelled word found in the
                yielded error data.
            include_word_line_index (bool): Includes the index of the caracter
                in which the mispelled word starts in their line (starting at
                index 0).
            include_line (bool): Includes the entire line where the mispelled
                word resides inside the content.
            include_text (bool): Includes the full text of the content in where
                the mispelled word resides.
            include_error_number (bool): Include the number of the error in
                yielded data. This could be useful to avoid the need of define
                a counter.
            include_near_misses (bool): Includes a list with the near misses
                for the mispelled word.

        Yields:
            dict: Dictionary with all the included data for each mispelled word.
        """
        self.errors = yield from parse_hunspell_output(
            self.filenames_contents,
            hunspell_spellcheck(
                quote_for_hunspell("\n".join(self.filenames_contents.values())),
                self.languages,
                personal_dicts=self.personal_dicts,
                encoding=self.encoding,
            ),
            looks_like_a_word=self.looks_like_a_word,
            include_filename=include_filename,
            include_line_number=include_line_number,
            include_word=include_word,
            include_word_line_index=include_word_line_index,
            include_line=include_line,
            include_text=include_text,
            include_error_number=include_error_number,
            include_near_misses=include_near_misses,
        )


def quote_for_hunspell(text):
    """Quote a paragraph so hunspell don't misinterpret it.

    Quoting Hunspell's manpage: "It is recommended that programmatic interfaces
    prefix every data line with an uparrow to protect themselves against future
    changes in hunspell."

    Args:
        text (str): Text to be quoted.

    Returns:
        str: Text quoted as Hunspell recommends.
    """
    response = []
    for line in text.splitlines():
        response.append(f"^{line}" if line else "")
    return "\n".join(response)


def parse_hunspell_output(
    filenames_contents,
    hunspell_output,
    looks_like_a_word=DEFAULT_LOOKS_LIKE_A_WORD,
    include_filename=True,
    include_line_number=True,
    include_word=True,
    include_word_line_index=True,
    include_line=False,
    include_text=False,
    include_error_number=False,
    include_near_misses=False,
):
    """Parse `hunspell -a` output."""
    locals_yielder = []

    _locals = locals()
    for possible_inclusion in ERROR_FIELDS:
        if _locals.get(f"include_{possible_inclusion}"):
            locals_yielder.append(possible_inclusion)

    error_number = 0
    checked_files = iter(filenames_contents.items())
    filename, text = next(checked_files)
    checked_lines = iter(text.split("\n"))
    line = next(checked_lines)
    line_number = 1

    for hunspell_line in hunspell_output.stdout.split("\n")[1:]:
        if not hunspell_line:
            try:
                line = next(checked_lines)
                line_number += 1
            except StopIteration:
                # next file
                try:
                    filename, text = next(checked_files)
                    checked_lines = iter(text.split("\n"))
                    line = next(checked_lines)
                    line_number = 1
                except StopIteration:
                    return error_number
            continue

        if hunspell_line[0] == "&":
            _, word, *mispell_data = hunspell_line.split()
            if include_word_line_index:
                word_line_index = int(mispell_data[1].rstrip(":")) - 1
            if include_near_misses:
                near_misses = [miss.rstrip(",") for miss in mispell_data[2:]]
            if looks_like_a_word(word):
                error_number += 1
                _locals = locals()
                yielded_content = {}
                for field in locals_yielder:
                    value = _locals.get(field)
                    if value:
                        yielded_content[field] = value
                    elif not isinstance(value, (str, list)):
                        yielded_content[field] = value
                yield yielded_content

    raise Unreachable(
        "This line shouldn't be reachable. Please, open an issue at"
        " https://github.com/mondeja/hunspellcheck"
    )  # pragma: no cover


def render_hunspell_word_error(
    data,
    fields=["filename", "word", "line_number", "word_line_index"],
    sep=":",
):
    """Renders a mispelled word data dictionary.

    This function allows a convenient way to render each mispelled word data
    dictionary as a string, that could be useful to print in the context of
    spell checkers command line interfaces.

    Args:
        data (dict): Mispelled word data, as it is yielded by the method
            :py:meth:`hunspellcheck.HunspellChecker.check`.
        fields (list): List of fields to include in the response.
        sep (str): Separator string between each field value.

    Returns:
        str: Mispelled word data as a string.
    """
    values = []
    for field in fields:
        value = data.get(field)
        if value is not None:
            values.append(str(value))
    return (sep).join(values)



"""CLI utilities writing spell checkers."""

import warnings

from hunspellcheck.cli.files import FilesOrGlobsAction
from hunspellcheck.cli.languages import create_hunspell_valid_dictionary_action
from hunspellcheck.cli.personal_dicts import PersonalDictionaryAction
from hunspellcheck.cli.version import DEFAULT_VERSION_TEMPLATE, render_version_template


def hunspellchecker_argument_parser(
    parser,
    version=False,
    version_prog=None,
    version_number=None,
    hunspell_version=True,
    ispell_version=True,
    version_template=DEFAULT_VERSION_TEMPLATE,
    version_template_context={},
    version_name_or_flags=["--version"],
    version_kwargs={},
    files=True,
    files_kwargs={},
    languages=True,
    languages_name_or_flags=["-l", "--language"],
    languages_kwargs={},
    negotiate_languages=True,
    personal_dicts=True,
    personal_dicts_name_or_flags=["-p", "--personal-dict"],
    personal_dicts_kwargs={},
    encoding=True,
    encoding_name_or_flags=["-i", "--input-encoding"],
    encoding_kwargs={},
    digits_are_words=True,
    digits_are_words_name_or_flags=["--digits-are-words"],
    digits_are_words_kwargs={},
    words_not_contain_digits=True,
    words_not_contain_digits_name_or_flags=["--words-not-contain-digits"],
    words_not_contain_digits_kwargs={},
    words_not_startswith_dash=True,
    words_not_startswith_dash_name_or_flags=["--words-not-startswith-dash"],
    words_not_startswith_dash_kwargs={},
    words_not_endswith_dash=True,
    words_not_endswith_dash_name_or_flags=["--words-not-endswith-dash"],
    words_not_endswith_dash_kwargs={},
    words_not_contain_dash=True,
    words_not_contain_dash_name_or_flags=["--words-not-contain-dash"],
    words_not_contain_dash_kwargs={},
    words_not_contain_two_upper=True,
    words_not_contain_two_upper_name_or_flags=["--words-not-contain-two-upper"],
    words_not_contain_two_upper_kwargs={},
    no_include_filename=True,
    no_include_filename_name_or_flags=["--no-include-filename"],
    no_include_filename_kwargs={},
    no_include_line_number=True,
    no_include_line_number_name_or_flags=["--no-include-line-number"],
    no_include_line_number_kwargs={},
    no_include_word=True,
    no_include_word_name_or_flags=["--no-include-word"],
    no_include_word_kwargs={},
    no_include_word_line_index=True,
    no_include_word_line_index_name_or_flags=["--no-include-word-line-index"],
    no_include_word_line_index_kwargs={},
    include_line=True,
    include_line_name_or_flags=["--include-line"],
    include_line_kwargs={},
    include_text=True,
    include_text_name_or_flags=["--include-text"],
    include_text_kwargs={},
    include_error_number=True,
    include_error_number_name_or_flags=["--include-error-number"],
    include_error_number_kwargs={},
    include_near_misses=True,
    include_near_misses_name_or_flags=["--include-near-misses"],
    include_near_misses_kwargs={},
):
    """Extends a :py:class:`argparse.ArgumentParser` instance adding
    spellchecking common parameters.

    By default will add next parameters:

    * A positional argument as a property named ``files`` inside the options
      namespace which takes multiple possible globs as inputs.
    * A required argument ``-l/--language`` that could be passed multiple times
      which take language dictionary names or filepaths. It will check if the
      passed language is recognized by Hunspell (or if is a dictionary file, if
      exists), and in case that not, will print a list with all available
      dictionaries.
    * An optional argument ``-p/--personal-dict`` that could be passed multiple
      times which takes a path to a file used to exclude certain words from
      being triggered as positives.
    * An optional argument ``-i/--input-encoding`` that should define the input
      content encoding.

    Args:
        version (bool): Include a convenient ``--version`` option that will
            print the version of the program, and optionally the installed
            versions of Hunspell and Ispell. See ``version_prog``,
            ``version_number``, ``hunspell_version`` and ``ispell_version``
            parameters below.
        version_prog (str): Name of the program shown along the version. If is
            not provided, will be taken from ``parser.prog`` property.
        version_number (str): Version of the program. See ``version_template``
            argument below for details about the formatting.
        hunspell_version (str): Include version of Hunspell in the version
            shown passing ``--version``.
        ispell_version (str): Include version of Ispell in the version shown
            passing ``--version``.
        version_template (str): Template for version rendering passed to a
            :py:class:`jinja2.Template` object that will be used to renderize
            the version string. By default, if ``version_number`` is provided,
            and ``hunspell_version`` and ``ispell_version`` are ``True``,
            it will render a string like
            ``"<version_prog> <X.Y.Z> - Hunspell <X.Y.Z> - Ispell <X.Y.Z>"``.
            The data for template rendering by default is compound by the next
            fields: ``version_prog``, ``version_number``, ``hunspell_version``
            and ``ispell_version``. If you want to pass other fields, include
            them in the argument ``version_template_context``.
        version_template_context (dict): Additional data to use in the version
            string rendering.
        version_name_or_flags (list, str): Flag name defined constructing the
            ``--version`` argument using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        version_kwargs (dict): Optional kwargs which override the default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            constructing the ``--version`` option.
        files (bool): Include the ``files`` positional argument inside the
            argument parser.
        files_kwargs (dict): Optional kwargs which override the default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            constructing the ``files`` positional argument.
        languages (bool): Include the ``-l/--language`` option inside the
            argument parser.
        languages_name_or_flags (list, str): Flag name defined constructing the
            ``-l/--language`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        languages_kwargs (dict): Optional kwargs which override the default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            constructing the ``-l/--language`` option.
        negotiate_languages (bool): Enables the language negotiation. If this
            is enabled and the CLI consumer passes a locale code instead of
            a full language name (for example `es` instead of `es_ES`),
            hunspellcheck will convert `es` to a territorialized language
            dictionary name available using the function
            :py:meth:`babel.core.Locale.negotiate`. If is disabled, a language
            dictionary passed as locale code like `es` will be considered
            invalid.
        personal_dicts (bool): Include the ``-p/--personal-dict`` option inside
            the argument parser.
        personal_dicts_name_or_flags (list, str): Flag name defined constructing
            the ``-p/--personal-dict`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        personal_dicts_kwargs (dict): Optional kwargs which override the default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            constructing the ``-p/--personal-dict`` option.
        encoding (bool): Include the ``-i/--input-encoding`` hunspell option
            inside the argument parser.
        encoding_name_or_flags (list, str): Flag name defined constructing
            the ``-i/--input-encoding`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        encoding_kwargs (dict): Optional kwargs which override the default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            building the ``-i/--input-encoding`` option.
        digits_are_words (bool): Include the option ``--digits-are-words`` to
            define if a value filled by digits will be considered a word for
            mispellchecking or not.
        digits_are_words_name_or_flags (list, str): Flag name defined
            constructing the ``--digits-are-words`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        digits_are_words_kwargs (dict): Optional kwargs which override default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            building the ``--digits-are-words` option.
        words_not_contain_digits (bool): Include the option
            ``--words-not-contain-digits`` which when passed in a CLI, the
            words that contain digits will be ignored mispellchecking errors.
        words_not_contain_digits_name_or_flags (list): Flag name defined
            constructing the ``--words-not-contain-digits`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        words_not_contain_digits_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--words-not-contain-digits` option.
        words_not_startswith_dash (bool): Include the option
            ``--words-not-startswith-dash`` which when passed in a CLI, the
            words starting with character ``"-" `` will be ignored
            mispellchecking errors.
        words_not_startswith_dash_name_or_flags (list): Flag name defined
            constructing the ``--words-not-startswith-dash`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        words_not_startswith_dash_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--words-not-startswith-dash` option.
        words_not_endswith_dash (bool): Include the option
            ``--words-not-endswith-dash`` which when passed in a CLI, the
            words ending with character ``"-" `` will be ignored
            mispellchecking errors.
        words_not_endswith_dash_name_or_flags (list): Flag name defined
            constructing the ``--words-not-endswith-dash`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        words_not_endswith_dash_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--words-not-endswith-dash` option.
        words_not_contain_dash (bool): Include the option
            ``--words-not-contain-dash`` which when passed in a CLI, the
            words containing character ``"-" `` will be ignored mispellchecking
            for possible errors.
        words_not_contain_dash_name_or_flags (list): Flag name defined
            constructing the ``--words-not-contain-dash`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        words_not_contain_dash_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--words-not-contain-dash` option.
        words_not_contain_two_upper (bool): Include the option
            ``--words-not-contain-two-upper`` which when passed in a CLI, the
            words containing two uppercase letters or mote will be ignored
            mispellchecking for possible errors.
        words_not_contain_two_upper_name_or_flags (list): Flag name defined
            constructing the ``--words-not-contain-two-upper`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        words_not_contain_two_upper_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--words-not-contain-two-upper` option.
        no_include_filename (bool): Include the option
            ``--no-include-filename`` which when passed in a CLI, the path to
            files in which mispelling errors are found are not shown in
            the output.
        no_include_filename_name_or_flags (list): Flag name defined
            constructing the ``--no-include-filename`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        no_include_filename_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--no-include-filename` option.
        no_include_line_number (bool): Include the option
            ``--no-include-line-number`` which when passed in a CLI, the number
            of lines in which mispelling errors are found are not shown in
            the output.
        no_include_line_number_name_or_flags (list): Flag name defined
            constructing the ``--no-include-line-number`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        no_include_line_number_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--no-include-line-number` option.
        no_include_word (bool): Include the option
            ``--no-include-word`` which when passed in a CLI, the words
            in which mispelling errors are found are not shown in the output.
        no_include_word_name_or_flags (list): Flag name defined
            constructing the ``--no-include-word`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        no_include_word_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--no-include-word` option.
        no_include_word_line_index (bool): Include the option
            ``--no-include-word-line-index`` which when passed in a CLI, the
            index of the mispelled words inside their lines in which mispelling
            errors are found are not shown in the output.
        no_include_word_line_index_name_or_flags (list): Flag name defined
            constructing the ``--no-include-word-line-index`` option using the
            method :py:meth:`argparse.ArgumentParser.add_argument`.
        no_include_word_line_index_kwargs (dict): Optional kwargs which
            override default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--no-include-word-line-index` option.
        include_line (bool): Include the option ``--include-line`` which when
            passed in a CLI, the line of the mispelled words in which
            mispelling errors are found are shown in the output.
        include_line_name_or_flags (list): Flag name defined constructing the
            ``--include-line`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        include_line_kwargs (dict): Optional kwargs which override default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            building the ``--include-line` option.
        include_text (bool): Include the option ``--include-text`` which when
            passed in a CLI, the text in which reside found mispelled words
            is shown in the output.
        include_text_name_or_flags (list): Flag name defined constructing the
            ``--include-text`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        include_text_kwargs (dict): Optional kwargs which override default
            kwargs passed to :py:meth:`argparse.ArgumentParser.add_argument`
            building the ``--include-text` option.
        include_error_number (bool): Include the option
            ``--include-error-number`` which when passed in a CLI, the number
            of each error is shown in the output.
        include_error_number_name_or_flags (list): Flag name defined building
            the ``--include-error-number`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        include_error_number_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--include-error-number` option.
        include_near_misses (bool): Include the option
            ``--include-near-misses`` which when passed in a CLI, some Hunspell
            suggestions will be shown for each mispelled word in the report.
        include_near_misses_name_or_flags (list): Flag name defined building
            the ``--include-near-misses`` option using the method
            :py:meth:`argparse.ArgumentParser.add_argument`.
        include_near_misses_kwargs (dict): Optional kwargs which override
            default kwargs passed to
            :py:meth:`argparse.ArgumentParser.add_argument` building the
            ``--include-near-misses` option.

    Examples:
        >>> import argparse
        >>>
        >>> parser = argparse.ArgumentParser()
        >>> hunspellchecker_argument_parser(
        ...     version=True,
        ...     version_number="1.0.0",
        ... )
        >>> opts = parser.parse_args(["--language", "es"])
        >>> print(opts)
        Namespace(languages=["es_ES"])
    """
    if version:
        version_string = render_version_template(
            version_template,
            version_template_context,
            version_prog=version_prog if version_prog is not None else parser.prog,
            version_number=version_number,
            hunspell_version=hunspell_version,
            ispell_version=ispell_version,
        )

        if version_string:
            _version_kwargs = {"action": "version", "version": version_string}
            _version_kwargs.update(version_kwargs)

            if isinstance(version_name_or_flags, str):
                version_name_or_flags = [version_name_or_flags]

            parser.add_argument(*version_name_or_flags, **_version_kwargs)
        else:
            readable_version_option = "/".join(version_name_or_flags)
            warnings.warn(
                f"'{readable_version_option}' option not added because version"
                " string is empty!"
            )

    if files:
        _files_kwargs = {
            "nargs": "*",
            "type": str,
            "dest": "files",
            "metavar": "FILES",
            "help": "Files and/or globs to check.",
            "action": FilesOrGlobsAction,
        }
        _files_kwargs.update(files_kwargs)
        parser.add_argument(**_files_kwargs)

    if languages:
        _languages_kwargs = {
            "type": str,
            "required": True,
            "metavar": "LANGUAGE",
            "dest": "languages",
            "help": "Language to check, you'll have to install the"
            " corresponding hunspell dictionary.",
        }

        _languages_kwargs["action"] = create_hunspell_valid_dictionary_action(
            negotiate_languages=negotiate_languages,
        )
        _languages_kwargs.update(languages_kwargs)

        if isinstance(languages_name_or_flags, str):
            languages_name_or_flags = [languages_name_or_flags]

        parser.add_argument(*languages_name_or_flags, **_languages_kwargs)

    if personal_dicts:
        _personal_dicts_kwargs = {
            "type": str,
            "required": False,
            "metavar": "PERSONAL_DICTIONARY",
            "dest": "personal_dicts",
            "help": "Additional dictionaries to extend the words to exclude.",
            "action": PersonalDictionaryAction,
            "nargs": 1,
            "default": None,
        }
        _personal_dicts_kwargs.update(personal_dicts_kwargs)

        if isinstance(personal_dicts_name_or_flags, str):
            personal_dicts_name_or_flags = [personal_dicts_name_or_flags]

        parser.add_argument(*personal_dicts_name_or_flags, **_personal_dicts_kwargs)

    if encoding:
        _encoding_kwargs = {
            "type": str,
            "required": False,
            "metavar": "ENCODING",
            "dest": "encoding",
            "help": "Input content encoding.",
            "action": "store",
            "default": None,
        }
        _encoding_kwargs.update(encoding_kwargs)

        if isinstance(encoding_name_or_flags, str):
            encoding_name_or_flags = [encoding_name_or_flags]

        parser.add_argument(*encoding_name_or_flags, **_encoding_kwargs)

    if digits_are_words:
        _digits_are_words_kwargs = {
            "action": "store_true",
            "required": False,
            "dest": "digits_are_words",
            "default": False,
            "help": (
                "Words compound only by digits will be considered words and"
                " will be checked for possible mispelling errors."
            ),
        }
        _digits_are_words_kwargs.update(digits_are_words_kwargs)

        if isinstance(digits_are_words_name_or_flags, str):
            digits_are_words_name_or_flags = [digits_are_words_name_or_flags]

        parser.add_argument(*digits_are_words_name_or_flags, **_digits_are_words_kwargs)

    if words_not_contain_digits:
        _words_not_contain_digits_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "words_can_contain_digits",
            "default": True,
            "help": (
                "Words that contain a digit inside will not be considered words"
                " and will be ignored checking possible mispelling errors."
            ),
        }
        _words_not_contain_digits_kwargs.update(words_not_contain_digits_kwargs)

        if isinstance(words_not_contain_digits_name_or_flags, str):
            words_not_contain_digits_name_or_flags = [
                words_not_contain_digits_name_or_flags
            ]

        parser.add_argument(
            *words_not_contain_digits_name_or_flags, **_words_not_contain_digits_kwargs
        )

    if words_not_startswith_dash:
        _words_not_startswith_dash_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "words_can_startswith_dash",
            "default": True,
            "help": (
                "Words starting with the character '-' will not be checked for"
                " possible mispelling errors."
            ),
        }
        _words_not_startswith_dash_kwargs.update(words_not_startswith_dash_kwargs)

        if isinstance(words_not_startswith_dash_name_or_flags, str):
            words_not_startswith_dash_name_or_flags = [
                words_not_startswith_dash_name_or_flags
            ]

        parser.add_argument(
            *words_not_startswith_dash_name_or_flags,
            **_words_not_startswith_dash_kwargs,
        )

    if words_not_endswith_dash:
        _words_not_endswith_dash_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "words_can_endswith_dash",
            "default": True,
            "help": (
                "Words ending with the character '-' will not be checked for"
                " possible mispelling errors."
            ),
        }
        _words_not_endswith_dash_kwargs.update(words_not_endswith_dash_kwargs)

        if isinstance(words_not_endswith_dash_name_or_flags, str):
            words_not_endswith_dash_name_or_flags = [
                words_not_endswith_dash_name_or_flags
            ]

        parser.add_argument(
            *words_not_endswith_dash_name_or_flags, **_words_not_endswith_dash_kwargs
        )

    if words_not_contain_dash:
        _words_not_contain_dash_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "words_can_contain_dash",
            "default": True,
            "help": (
                "Words containing the character '-' will not be checked for"
                " possible mispelling errors."
            ),
        }
        _words_not_contain_dash_kwargs.update(words_not_contain_dash_kwargs)

        if isinstance(words_not_contain_dash_name_or_flags, str):
            words_not_contain_dash_name_or_flags = [
                words_not_contain_dash_name_or_flags
            ]

        parser.add_argument(
            *words_not_contain_dash_name_or_flags, **_words_not_contain_dash_kwargs
        )

    if words_not_contain_two_upper:
        _words_not_contain_two_upper_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "words_can_contain_two_upper",
            "default": True,
            "help": (
                "Words containing two or more uppercase letters will not be"
                " checked for possible mispelling errors."
            ),
        }
        _words_not_contain_two_upper_kwargs.update(words_not_contain_two_upper_kwargs)

        if isinstance(words_not_contain_two_upper_name_or_flags, str):
            words_not_contain_two_upper_name_or_flags = [
                words_not_contain_two_upper_name_or_flags
            ]

        parser.add_argument(
            *words_not_contain_two_upper_name_or_flags,
            **_words_not_contain_two_upper_kwargs,
        )

    if no_include_filename:
        _no_include_filename_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "include_filename",
            "default": True,
            "help": "Filenames will not be included in spellchecking errors report.",
        }
        _no_include_filename_kwargs.update(no_include_filename_kwargs)

        if isinstance(no_include_filename_name_or_flags, str):
            no_include_filename_name_or_flags = [no_include_filename_name_or_flags]

        parser.add_argument(
            *no_include_filename_name_or_flags, **_no_include_filename_kwargs
        )

    if no_include_line_number:
        _no_include_line_number_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "include_line_number",
            "default": True,
            "help": "Line numbers will not be included in spellchecking errors report.",
        }
        _no_include_line_number_kwargs.update(no_include_line_number_kwargs)

        if isinstance(no_include_line_number_name_or_flags, str):
            no_include_line_number_name_or_flags = [
                no_include_line_number_name_or_flags
            ]

        parser.add_argument(
            *no_include_line_number_name_or_flags, **_no_include_line_number_kwargs
        )

    if no_include_word:
        _no_include_word_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "include_word",
            "default": True,
            "help": "Words will not be included in spellchecking errors report.",
        }
        _no_include_word_kwargs.update(no_include_word_kwargs)

        if isinstance(no_include_word_name_or_flags, str):
            no_include_word_name_or_flags = [no_include_word_name_or_flags]

        parser.add_argument(*no_include_word_name_or_flags, **_no_include_word_kwargs)

    if no_include_word_line_index:
        _no_include_word_line_index_kwargs = {
            "action": "store_false",
            "required": False,
            "dest": "include_word_line_index",
            "default": True,
            "help": (
                "Index of mispelled words inside their lines will not be"
                " included in spellchecking errors report."
            ),
        }
        _no_include_word_line_index_kwargs.update(no_include_word_line_index_kwargs)

        if isinstance(no_include_word_line_index_name_or_flags, str):
            no_include_word_line_index_name_or_flags = [
                no_include_word_line_index_name_or_flags
            ]

        parser.add_argument(
            *no_include_word_line_index_name_or_flags,
            **_no_include_word_line_index_kwargs,
        )

    if include_line:
        _include_line_kwargs = {
            "action": "store_true",
            "required": False,
            "dest": "include_line",
            "default": False,
            "help": (
                "Entire lines inside mispelled words are found will be shown in"
                " the report."
            ),
        }
        _include_line_kwargs.update(include_line_kwargs)

        if isinstance(include_line_name_or_flags, str):
            include_line_name_or_flags = [include_line_name_or_flags]

        parser.add_argument(*include_line_name_or_flags, **_include_line_kwargs)

    if include_text:
        _include_text_kwargs = {
            "action": "store_true",
            "required": False,
            "dest": "include_text",
            "default": False,
            "help": (
                "Entire texts inside mispelled words are found will be shown"
                " in the report."
            ),
        }
        _include_text_kwargs.update(include_text_kwargs)

        if isinstance(include_text_name_or_flags, str):
            include_text_name_or_flags = [include_text_name_or_flags]

        parser.add_argument(*include_text_name_or_flags, **_include_text_kwargs)

    if include_error_number:
        _include_error_number_kwargs = {
            "action": "store_true",
            "required": False,
            "dest": "include_error_number",
            "default": False,
            "help": ("A counter for each mispelled word will be shown in the report."),
        }
        _include_error_number_kwargs.update(include_error_number_kwargs)

        if isinstance(include_error_number_name_or_flags, str):
            include_error_number_name_or_flags = [include_error_number_name_or_flags]

        parser.add_argument(
            *include_error_number_name_or_flags, **_include_error_number_kwargs
        )

    if include_near_misses:
        _include_near_misses_kwargs = {
            "action": "store_true",
            "required": False,
            "dest": "include_near_misses",
            "default": False,
            "help": (
                "Some Hunspell suggestions for each mispelled word will be"
                " shown in the report."
            ),
        }
        _include_near_misses_kwargs.update(include_near_misses_kwargs)

        if isinstance(include_near_misses_name_or_flags, str):
            include_near_misses_name_or_flags = [include_near_misses_name_or_flags]

        parser.add_argument(
            *include_near_misses_name_or_flags, **_include_near_misses_kwargs
        )



"""Exceptons module of hunspellcheck."""


class HunspellCheckError(Exception):
    """All exceptions from this module inherit from this one."""


class Unreachable(HunspellCheckError):
    """The code encontered a state that should be unreachable."""


class InvalidLanguageDictionaryError(HunspellCheckError, ValueError):
    """Invalid language dictionary name."""



"""Utilities about Hunspell dictionaries."""

import os
import subprocess
import sys

from babel import Locale

from hunspellcheck.exceptions import InvalidLanguageDictionaryError


def gen_available_dictionaries(full_paths=False):
    """Generates the available dictionaries contained inside the search paths
    configured by hunspell.

    These dictionaries can be used without specify the full path to their
    location in the system calling hunspell, only their name is needed.

    Args:
        full_paths (bool): Yield complete paths to dictionaries (``True``) or
            their names only (``False``).

    Yields:
        str: Dictionary names (locale with territory).
    """
    previous_env_lang = os.environ.get("LANG", None)
    os.environ["LANG"] = "C"

    output = subprocess.run(
        ["hunspell", "-D"],
        stderr=subprocess.PIPE,
        universal_newlines=True,
    )

    if previous_env_lang is None:
        del os.environ["LANG"]
    else:
        os.environ["LANG"] = previous_env_lang

    _inside_available_dictionaries = False
    for line in output.stderr.splitlines():
        if _inside_available_dictionaries:
            yield line if full_paths else os.path.basename(line)
        elif line.startswith("AVAILABLE DICTIONARIES"):
            _inside_available_dictionaries = True


def list_available_dictionaries(full_paths=False):
    """Convenient wrapper around the generator
    :py:func:`hunspellcheck.gen_available_dictionaries`
    which returns the dictionary names in a list.

    Args:
        full_paths (bool): Print complete paths to dictionaries (``True``) or
            their names only (``False``).

    Returns:
        list: Available installed dictionaries.
    """
    return list(gen_available_dictionaries(full_paths=full_paths))


def print_available_dictionaries(sort=True, stream=sys.stdout, full_paths=False):
    """Prints into an stream the available hunspell dictionaries.

    By default are printed to the standard output of the system (STDOUT).

    Args:
        sort (bool): Indicates if the dictionaries will be printed in
            alphabetical order.
        stream (object): Stream to which the dictionaries will be printed.
            Must be any object that accepts a `write` method.
        full_paths (bool): Print complete paths to dictionaries (``True``) or
            their names only (``False``).
    """
    if sort:
        dictionaries_iter = sorted(list_available_dictionaries(full_paths=full_paths))
    else:
        dictionaries_iter = gen_available_dictionaries(full_paths=full_paths)

    for dictname in dictionaries_iter:
        stream.write(f"{dictname}\n")


def gen_available_dictionaries_with_langcodes(sort=True, full_paths=False):
    """Generates all available dictionaries installed along with their
    locale names (without territories).

    For example, if `es_ES` is installed, `es` also will be included
    in the response.

    Args:
        sort (bool): Sort languages alfabetically.

    Yields:
        str: Locale or dictionary names (locale with territory).
    """
    dictionaries_iter = gen_available_dictionaries(full_paths=full_paths)
    if sort:
        dictionaries_iter = sorted(dictionaries_iter)
    unique_locales = []
    for dictname in dictionaries_iter:
        if "_" in dictname:
            locale = dictname.split("_")[0]
            if locale not in unique_locales:
                unique_locales.append(locale)
                yield locale
        yield dictname


def is_valid_dictionary_language(dictionary_name, negotiate_languages=False):
    """Check if a dictionary name is a valid dictionary installed
    for your Hunspell version.

    Args:
        dictionary_name (str): Dictionary language.
        negotiate_languages (bool): Enable language negotiation from locale
            name to territory.

    Returns:
        tuple: Has 3 values:

        - The first value is a boolean and indicates if the language is valid.
        - The second value is the dictionary language name, which could be
          changed from the input is language negotation is enabled.
        - The third value is a list with all available dictionaries.
    """
    available_dictionaries = list_available_dictionaries()
    if dictionary_name not in available_dictionaries:
        if negotiate_languages:
            dictionary_name = str(
                Locale.negotiate([dictionary_name], available_dictionaries)
            )
        else:
            return (False, None, available_dictionaries)
    return (True, dictionary_name, available_dictionaries)


def is_valid_dictionary_language_or_filename(value, negotiate_languages=False):
    """Returns if a value is a valid dictionary language name or an existent
    file defined by their path.

    Args:
        value (str): Dictionary language or filepath.
        negotiate_languages (bool): Enable language negotiation from locale
            name to territory.

    Returns:
        bool: Indicates if is a valid dictionary supported by Hunspell.
    """
    if os.path.isfile(value):
        return True
    is_valid, *_ = is_valid_dictionary_language(
        value,
        negotiate_languages=negotiate_languages,
    )
    return is_valid


def assert_is_valid_dictionary_language_or_filename(
    value,
    negotiate_languages=False,
):
    """Asserts if a value is a valid dictionary language name or an existent
    file defined by their path. If is not, raises an
    :py:class:`hunspellcheck.InvalidLanguageDictionaryError`.

    Args:
        value (str, list): Dictionary language/s or filepath/s.
        negotiate_languages (bool): Enable language negotiation from locale
            name to territory.
    """
    if isinstance(value, str):
        if not is_valid_dictionary_language_or_filename(
            value,
            negotiate_languages=negotiate_languages,
        ):
            raise InvalidLanguageDictionaryError(value)
    else:
        for language in value:
            if not is_valid_dictionary_language_or_filename(
                language,
                negotiate_languages=negotiate_languages,
            ):
                raise InvalidLanguageDictionaryError(value)



"""Language option related stuff for hunspellcheck CLI utilities."""

import argparse
import os

from hunspellcheck.hunspell.dictionaries import (
    gen_available_dictionaries_with_langcodes,
    is_valid_dictionary_language,
)


def create_hunspell_valid_dictionary_action(negotiate_languages=True):
    class HunspellDictionaryNegotiatorAction(argparse._AppendAction):
        """This action allows to redirect a language like 'es' or a dictionary
        filepath to a valid language dictionary supported by hunspell.
        """

        def __call__(self, parser, namespace, values, option_string=None):
            if os.path.isfile(values):
                values = values.rstrip(".dic").rstrip(".aff")
            else:
                (
                    is_valid,
                    dictionary_language,
                    available_dictionaries,
                ) = is_valid_dictionary_language(
                    values,
                    negotiate_languages=negotiate_languages,
                )
                if is_valid:
                    values = dictionary_language

                self.choices = (
                    available_dictionaries
                    if not negotiate_languages
                    else list(gen_available_dictionaries_with_langcodes())
                )
                self.choices.append("<dictionary filepath>")

            # check value manually (seems to me like a hack, but it works)
            parser._check_value(self, values)
            super().__call__(parser, namespace, values, option_string=option_string)

    return HunspellDictionaryNegotiatorAction



import string  # noqa: F401
import unicodedata  # noqa: F401


def looks_like_a_word_creator(
    digits_are_words=False,
    words_can_contain_digits=True,
    words_can_startswith_dash=True,
    words_can_endswith_dash=True,
    words_can_contain_dash=True,
    words_can_contain_two_upper=True,
):
    """Generates dinamically the function ``look_like_a_word`` use to clean the
    words that must not be checked for mispelling errors.

    Args:
        digits_are_words (bool): If ``False``, values with all characters as
            digits will not be considered words.
        words_can_contain_digits (bool): If ``False``, values with at least one
            digit character will not be considered words.
        words_can_startswith_dash (bool): If ``False``, values starting with the ``-``
            character will not be considered words.
        words_can_endswith_dash (bool): If ``False``, values ending with the ``-``
            character will not be considered words.
        words_can_contain_dash (bool): If ``False``, values containing the ``-``
            character will not be considered words.
        words_can_contain_two_upper (bool): If ``False``, values which
            contain at least two uppercase like CPython will not be considered
            words and will not be checking for possible mispellings.

    Returns:
        function: Function that takes a possible word as a parameter and
            returns if that value is considered a word. This function can be
            passed to :py:class:`hunspellcheck.spellchecker.HunspellChecker`.
    """
    conditions = [
        (digits_are_words, "all(char in string.digits for char in text)"),
        (words_can_contain_digits, "any(digit in text for digit in string.digits)"),
        (words_can_startswith_dash, 'text.startswith("-")'),
        (words_can_endswith_dash, 'text.endswith("-")'),
        (words_can_contain_dash, '"-" in text'),
        (
            words_can_contain_two_upper,
            'len([c for c in text if unicodedata.category(c) == "Lu"]) > 1',
        ),
    ]

    function_conditional = "if not text or "

    for i, (argument, condition) in enumerate(conditions):
        if argument is False:
            if not function_conditional.endswith(" or "):
                function_conditional += " or "
            function_conditional += condition
    function_conditional += ":\n        return False\n    "

    function_definition = f"""def looks_like_a_word(text):
    {function_conditional}return True"""

    code = compile(function_definition, "test", "exec")
    exec(code)
    return locals()["looks_like_a_word"]
